<?php

    namespace FederationLib;

    use Exception;
    use FederationLib\Classes\Configuration;
    use FederationLib\Enums\Standard\ErrorCodes;
    use FederationLib\Enums\Standard\Methods;
    use FederationLib\Exceptions\DatabaseException;
    use FederationLib\Exceptions\Standard\AccessDeniedException;
    use FederationLib\Exceptions\Standard\ClientNotFoundException;
    use FederationLib\Exceptions\Standard\InternalServerException;
    use FederationLib\Exceptions\Standard\InvalidClientDescriptionException;
    use FederationLib\Exceptions\Standard\InvalidClientNameException;
    use FederationLib\Managers\ClientManager;
    use FederationLib\Objects\ClientRecord;
    use FederationLib\Objects\ResolvedIdentity;
    use FederationLib\Objects\Standard\ClientIdentity;
    use TamerLib\Enums\TamerMode;
    use TamerLib\tm;
    use Throwable;

    class FederationLib
    {
        /**
         * @var ClientManager
         */
        private ClientManager $client_manager;

        /**
         * FederationLib constructor.
         */
        public function __construct()
        {
            $this->client_manager = new ClientManager($this);
        }

        /**
         * Registers functions to the TamerLib instance, if applicable
         *
         * @return void
         */
        public function registerFunctions(): void
        {
            if(tm::getMode() !== TamerMode::WORKER)
            {
                return;
            }

            $this->client_manager->registerFunctions();
        }

        /**
         * @return string
         */
        public static function getSubprocessPath(): string
        {
            return __DIR__ . DIRECTORY_SEPARATOR . 'subproc';
        }

        /**
         * Resolves the permission role from the given identity and attempts to check if the identity has the
         * required permission to perform the given method
         *
         * @param ClientIdentity|null $identity
         * @throws AccessDeniedException
         * @throws ClientNotFoundException
         * @throws InternalServerException
         * @return ResolvedIdentity
         */
        private function resolveIdentity(?ClientIdentity $identity): ResolvedIdentity
        {
            if($identity === null)
            {
                return new ResolvedIdentity(null, null, true);
            }

            $get_client = tm::do('client_getClient', [$identity->getClientUuid()]);
            $peer = null;

            try
            {
                $client = tm::waitFor($get_client);
                tm::dof('client_updateLastSeen');
            }
            catch(ClientNotFoundException $e)
            {
                tm::clear();
                throw new ClientNotFoundException('Invalid client UUID or client does not exist', $e);
            }
            catch(Exception|Throwable $e)
            {
                tm::clear();
                throw new InternalServerException('There was an error while trying to access the client', $e);
            }

            return new ResolvedIdentity($client, $peer);
        }

        /**
         * Checks if the given identity has the required permission to perform the given method
         *
         * @param string $method
         * @param ResolvedIdentity $resolved_identity
         * @return bool
         */
        private function checkPermission(string $method, ResolvedIdentity $resolved_identity): bool
        {
            return $resolved_identity->getPermissionRole() <= Configuration::getMethodPermission($method);
        }

        /**
         * Pings the client
         *
         * @param ClientIdentity|null $identity
         * @return bool
         * @throws AccessDeniedException
         * @throws ClientNotFoundException
         * @throws InternalServerException
         */
        public function ping(?ClientIdentity $identity): bool
        {
            if(!$this->checkPermission(Methods::PING, $this->resolveIdentity($identity)))
            {
                throw new Exceptions\Standard\AccessDeniedException('You do not have permission to perform this action');
            }

            return true;
        }

        /**
         * @param ClientIdentity|null $identity
         * @return string
         * @throws AccessDeniedException
         * @throws ClientNotFoundException
         * @throws InternalServerException
         */
        public function whoami(?ClientIdentity $identity): string
        {
            $resolved_identity = $this->resolveIdentity($identity);

            if(!$this->checkPermission(Methods::WHOAMI, $resolved_identity))
            {
                throw new Exceptions\Standard\AccessDeniedException('You do not have permission to perform this action');
            }

            if($resolved_identity->getPeer() !== null)
            {
                return $resolved_identity->getPeer()->getFederatedAddress();
            }

            if($resolved_identity->getClient() !== null)
            {
                return $resolved_identity->getClient()->getUuid();
            }

            return 'root';
        }

        /**
         * Registers a new client into the database
         *
         * @param ClientIdentity|null $identity
         * @param string|null $name
         * @param string|null $description
         * @return string
         * @throws AccessDeniedException
         * @throws ClientNotFoundException
         * @throws DatabaseException
         * @throws InternalServerException
         * @throws InvalidClientDescriptionException
         * @throws InvalidClientNameException
         */
        public function createClient(?ClientIdentity $identity, ?string $name=null, ?string $description=null): string
        {
            if(!$this->checkPermission(Methods::CREATE_CLIENT, $this->resolveIdentity($identity)))
            {
                throw new Exceptions\Standard\AccessDeniedException('You do not have sufficient permission to create a client');
            }

            try
            {
                return $this->client_manager->registerClient($name, $description);
            }
            catch(Exception $e)
            {
                if(in_array($e->getCode(), ErrorCodes::ALL, true))
                {
                    throw $e;
                }

                throw new Exceptions\Standard\InternalServerException('There was an error while creating the client', $e);
            }
        }

        /**
         * Returns an existing client from the database
         *
         * @param ClientIdentity|null $identity
         * @param string|ClientRecord $client_uuid
         * @return Objects\Standard\Client
         * @throws ClientNotFoundException
         * @throws DatabaseException
         * @throws InternalServerException
         * @throws AccessDeniedException
         */
        public function getClient(?ClientIdentity $identity, string|ClientRecord $client_uuid): Objects\Standard\Client
        {
            if(!$this->checkPermission(Methods::GET_CLIENT, $this->resolveIdentity($identity)))
            {
                throw new Exceptions\Standard\AccessDeniedException('You do not have sufficient permission to fetch a client from the database');
            }

            try
            {
                // Return the standard client object
                return new Objects\Standard\Client($this->client_manager->getClient($client_uuid));
            }
            catch(Exception $e)
            {
                if(in_array($e->getCode(), ErrorCodes::ALL, true))
                {
                    throw $e;
                }

                throw new Exceptions\Standard\InternalServerException('There was an error while getting the client', $e);
            }
        }

        /**
         * Updates the name of an existing client, return True if successful
         *
         * @param ClientIdentity|null $identity
         * @param string $client_uuid
         * @param string $new_name
         * @return bool
         * @throws AccessDeniedException
         * @throws ClientNotFoundException
         * @throws DatabaseException
         * @throws InternalServerException
         * @throws InvalidClientNameException
         */
        public function updateClientName(?ClientIdentity $identity, string $client_uuid, ?string $new_name): bool
        {
            if(!$this->checkPermission(Methods::UPDATE_CLIENT_NAME, $this->resolveIdentity($identity)))
            {
                throw new Exceptions\Standard\AccessDeniedException('You do not have sufficient permission to update the name of a client');
            }

            try
            {
                $this->client_manager->changeClientName($client_uuid, $new_name);
            }
            catch(Exception $e)
            {
                if(in_array($e->getCode(), ErrorCodes::ALL, true))
                {
                    throw $e;
                }

                throw new Exceptions\Standard\InternalServerException('There was an error while updating the client name', $e);
            }

            return true;
        }
    }