<?php

    namespace FederationLib\Managers;

    use Doctrine\DBAL\ParameterType;
    use Exception;
    use FederationLib\Classes\Configuration;
    use FederationLib\Classes\Database;
    use FederationLib\Classes\Validate;
    use FederationLib\Enums\DatabaseTables;
    use FederationLib\Enums\Misc;
    use FederationLib\Exceptions\DatabaseException;
    use FederationLib\Exceptions\PeerAssociationNotFoundException;
    use FederationLib\Exceptions\Standard\InvalidPeerAssociationTypeException;
    use FederationLib\FederationLib;
    use FederationLib\Objects\ClientRecord;
    use FederationLib\Objects\ParsedFederatedAddress;
    use FederationLib\Objects\PeerAssociationRecord;
    use LogLib\Log;

    class AssociationManager
    {
        /**
         * @var FederationLib
         */
        private FederationLib $federationLib;

        /**
         * AssociationManager constructor.
         *
         * @param FederationLib $federationLib
         */
        public function __construct(FederationLib $federationLib)
        {
            $this->federationLib = $federationLib;
        }

        /**
         * Associates a child peer with a parent peer.
         *
         * @param ClientRecord|string $client_uuid
         * @param ParsedFederatedAddress|string $parent
         * @param ParsedFederatedAddress|string $child
         * @param string $type
         * @return void
         * @throws DatabaseException
         * @throws InvalidPeerAssociationTypeException
         */
        public function associate(ClientRecord|string $client_uuid, ParsedFederatedAddress|string $parent, ParsedFederatedAddress|string $child, string $type): void
        {
            if(!Validate::peerAssociationType($type))
            {
                throw new InvalidPeerAssociationTypeException(sprintf('Invalid peer association type: %s', $type));
            }

            if($client_uuid instanceof ClientRecord)
            {
                $client_uuid = $client_uuid->getUuid();
            }

            if(is_string($parent))
            {
                $parent = new ParsedFederatedAddress($parent);
            }

            if(is_string($child))
            {
                $child = new ParsedFederatedAddress($child);
            }

            try
            {
                // Try updating the association if it already exists, but only if the association type is different
                $peer_association = $this->getAssociation($parent, $child);
                if($peer_association->getAssociationType() === $type)
                {
                    return;
                }
                $this->updateAssociation($client_uuid, $parent, $child, $type);
                return;
            }
            catch(PeerAssociationNotFoundException $e)
            {
                // This is fine, we'll just create a new association
                unset($e);
            }

            $qb = Database::getConnection()->createQueryBuilder();

            $qb->insert(DatabaseTables::PEERS_ASSOCIATIONS);
            $qb->setValue('child_peer', $qb->createNamedParameter($child->getAddress()));
            $qb->setValue('parent_peer', $qb->createNamedParameter($parent->getAddress()));
            $qb->setValue('association_type', $qb->createNamedParameter($type));
            $qb->setValue('client_uuid', $qb->createNamedParameter($client_uuid));
            $qb->setValue('timestamp', $qb->createNamedParameter(time(), ParameterType::INTEGER));

            try
            {
                $qb->executeStatement();
            }
            catch(Exception $e)
            {
                throw new DatabaseException('Failed to register peer association: ' . $e->getMessage(), $e);
            }

            Log::info(Misc::FEDERATIONLIB, sprintf('Associated %s with %s as %s', $child->getAddress(), $parent->getAddress(), $type));
        }

        /**
         * Returns the association record for the child peer and parent peer.
         *
         * @param ParsedFederatedAddress|string $parent
         * @param ParsedFederatedAddress|string $child
         * @return PeerAssociationRecord
         * @throws DatabaseException
         * @throws PeerAssociationNotFoundException
         */
        public function getAssociation(ParsedFederatedAddress|string $parent, ParsedFederatedAddress|string $child): PeerAssociationRecord
        {
            if(is_string($parent))
            {
                $parent = new ParsedFederatedAddress($parent);
            }

            if(is_string($child))
            {
                $child = new ParsedFederatedAddress($child);
            }

            $cache_key = sprintf('peer_association_%s_%s', $parent->getAddress(), $child->getAddress());
            if(Configuration::isCacheSystemEnabled() && Configuration::getObjectCacheEnabled('peer_association_objects'))
            {
                try
                {
                    $redis = RedisConnectionManager::getConnectionFromConfig('peer_association_objects');

                    if($redis->exists($cache_key))
                    {
                        $peer_association = PeerAssociationRecord::fromArray($redis->hGetAll($cache_key));
                        Log::debug(Misc::FEDERATIONLIB, sprintf('Loaded peer association object %s from cache', $cache_key));
                        return $peer_association;
                    }
                }
                catch(Exception $e)
                {
                    Log::warning(Misc::FEDERATIONLIB, sprintf('Failed to load peer association %s from cache: %s', $cache_key, $e->getMessage()));
                }
            }

            $qb = Database::getConnection()->createQueryBuilder();
            $qb->select(
                'child_peer',
                'parent_peer',
                'association_type',
                'client_uuid',
                'timestamp'
            );

            $qb->from(DatabaseTables::PEERS_ASSOCIATIONS);
            $qb->where('child_peer = :child_peer');
            $qb->andWhere('parent_peer = :parent_peer');
            $qb->setParameter('child_peer', $child->getAddress());
            $qb->setParameter('parent_peer', $parent->getAddress());
            $qb->setMaxResults(1);

            try
            {
                $result = $qb->executeQuery();

                if($result->rowCount() === 0)
                {
                    throw new PeerAssociationNotFoundException(sprintf('Peer association not found: %s -> %s', $child->getAddress(), $parent->getAddress()));
                }

                $peer_association = PeerAssociationRecord::fromArray($result->fetchAssociative());
            }
            catch(PeerAssociationNotFoundException $e)
            {
                throw $e;
            }
            catch(Exception $e)
            {
                throw new DatabaseException('Failed to get peer association: ' . $e->getMessage(), $e);
            }

            if(Configuration::isCacheSystemEnabled() && Configuration::getObjectCacheEnabled('peer_association_objects'))
            {
                try
                {
                    if(!isset($redis))
                    {
                        $redis = RedisConnectionManager::getConnectionFromConfig('peer_association_objects');
                    }

                    $redis->hMSet($cache_key, $peer_association->toArray());

                    if(Configuration::getObjectCacheTtl('peer_association_objects') > 0)
                    {
                        $redis->expire($cache_key, Configuration::getObjectCacheTtl('peer_association_objects'));
                    }

                    Log::debug(Misc::FEDERATIONLIB, sprintf('Cached peer association object %s', $cache_key));
                }
                catch(Exception $e)
                {
                    Log::warning(Misc::FEDERATIONLIB, sprintf('Failed to cache peer association %s: %s', $cache_key, $e->getMessage()));
                }
            }

            return $peer_association;
        }

        /**
         * Updates the association record for the child peer and parent peer.
         * If the type remains the same, nothing will be done.
         *
         * @param ClientRecord|string $client_uuid
         * @param ParsedFederatedAddress|string $parent
         * @param ParsedFederatedAddress|string $child
         * @param string $type
         * @return void
         * @throws DatabaseException
         * @throws InvalidPeerAssociationTypeException
         * @throws PeerAssociationNotFoundException
         */
        public function updateAssociation(ClientRecord|string $client_uuid, ParsedFederatedAddress|string $parent, ParsedFederatedAddress|string $child, string $type): void
        {
            if(!Validate::peerAssociationType($type))
            {
                throw new InvalidPeerAssociationTypeException(sprintf('Invalid peer association type: %s', $type));
            }

            if($client_uuid instanceof ClientRecord)
            {
                $client_uuid = $client_uuid->getUuid();
            }

            if(is_string($parent))
            {
                $parent = new ParsedFederatedAddress($parent);
            }

            if(is_string($child))
            {
                $child = new ParsedFederatedAddress($child);
            }

            $peer_association = $this->getAssociation($parent, $child);

            if($peer_association->getAssociationType() === $type)
            {
                return;
            }

            $qb = Database::getConnection()->createQueryBuilder();

            $qb->update(DatabaseTables::PEERS_ASSOCIATIONS);
            $qb->set('association_type', $qb->createNamedParameter($type));
            $qb->set('client_uuid', $qb->createNamedParameter($client_uuid));
            $qb->set('timestamp', $qb->createNamedParameter(time(), ParameterType::INTEGER));
            $qb->where('child_peer = :child_peer');
            $qb->andWhere('parent_peer = :parent_peer');
            $qb->setParameter('child_peer', $child->getAddress());
            $qb->setParameter('parent_peer', $parent->getAddress());

            try
            {
                $qb->executeStatement();
            }
            catch(Exception $e)
            {
                throw new DatabaseException('Failed to update peer association: ' . $e->getMessage(), $e);
            }

            $cache_key = sprintf('peer_association_%s_%s', $parent->getAddress(), $child->getAddress());

            if(Configuration::isCacheSystemEnabled() && Configuration::getObjectCacheEnabled('peer_association_objects'))
            {
                try
                {
                    $redis = RedisConnectionManager::getConnectionFromConfig('peer_association_objects');

                    if($redis->exists($cache_key))
                    {
                        $redis->hMSet($cache_key, [
                            'association_type' => $type,
                            'client_uuid' => $client_uuid,
                            'timestamp' => time()
                        ]);

                        if(Configuration::getObjectCacheTtl('peer_association_objects') > 0)
                        {
                            $redis->expire($cache_key, Configuration::getObjectCacheTtl('peer_association_objects'));
                        }

                        Log::debug(Misc::FEDERATIONLIB, sprintf('Updated cached peer association object %s', $cache_key));
                    }
                }
                catch(Exception $e)
                {
                    Log::warning(Misc::FEDERATIONLIB, sprintf('Failed to update cached peer association %s: %s', $cache_key, $e->getMessage()));
                }
            }

            Log::debug(Misc::FEDERATIONLIB, sprintf('Updated peer association %s -> %s: %s', $child->getAddress(), $parent->getAddress(), $type));
        }
    }