<?php

    /** @noinspection PhpUnused */
    /** @noinspection PhpMissingFieldTypeInspection */

    namespace FederationLib\Objects\Standard\PeerMetadata;

    use FederationLib\Classes\Validate;
    use FederationLib\Enums\Standard\UserPeerType;
    use FederationLib\Exceptions\Standard\InvalidPeerMetadataException;
    use FederationLib\Interfaces\PeerMetadataInterface;
    use FederationLib\Objects\ParsedFederatedAddress;

    class TelegramUserMetadata implements PeerMetadataInterface
    {
        /**
         * @var int
         */
        private $id;

        /**
         * @var bool
         */
        private $is_bot;

        /**
         * @var string
         */
        private $first_name;

        /**
         * @var string|null
         */
        private $last_name;

        /**
         * @var string|null
         */
        private $username;

        /**
         * @var string|null
         */
        private $language_code;

        /**
         * @var bool
         */
        private $is_premium;

        /**
         * @var int|null
         */
        private $updated_timestamp;

        /**
         * @var string|null
         */
        private $updated_client;

        /**
         * Validates the given metadata and throws an exception if the metadata is invalid
         *
         * @throws InvalidPeerMetadataException
         */
        public function validate(): void
        {
            $required_properties = [
                'id' => 'integer',
                'is_bot' => 'boolean',
                'first_name' => 'string',
                'is_premium' => 'boolean',
            ];

            $optional_properties = [
                'last_name' => 'string',
                'username' => 'string',
                'language_code' => 'string',
            ];

            Validate::metadata($this->toArray(), $required_properties, $optional_properties);
        }

        /**
         * Returns True if the peer is a bot, False otherwise
         *
         * @return bool
         */
        public function isAutomated(): bool
        {
            return $this->is_bot;
        }

        /**
         * Returns the platform-specific ID of the peer
         *
         * @return string
         */
        public function getPlatformId(): string
        {
            return $this->id;
        }

        /**
         * Returns the standard federated address of the peer
         * Format: telegram.user:<id>
         *
         * @return ParsedFederatedAddress
         */
        public function getFederatedAddress(): ParsedFederatedAddress
        {
            return new ParsedFederatedAddress(sprintf('%s:%s', UserPeerType::TELEGRAM_USER, $this->id));
        }

        /**
         * Returns the Client UUID that last updated the record
         *
         * @return ?string
         */
        public function getUpdatedClient(): ?string
        {
            return $this->updated_client;
        }

        /**
         * Sets the client UUID that last updated the record
         *
         * @param string|null $client_uuid
         * @return void
         */
        public function setUpdatedClient(?string $client_uuid): void
        {
            $this->updated_client = $client_uuid;
        }

        /**
         * Returns the Timestamp for when the record was last updated
         *
         * @return int
         */
        public function getLastUpdatedTimestamp(): int
        {
            return $this->updated_timestamp;
        }

        /**
         * Sets the Timestamp for when the record was last updated
         *
         * @param int|null $timestamp
         * @return void
         */
        public function setLastUpdatedTimestamp(?int $timestamp): void
        {
            $this->updated_timestamp = $timestamp;
        }

        /**
         * Unique identifier for this user or bot. This number may have more than 32 significant bits and some
         * programming languages may have difficulty/silent defects in interpreting it. But it has at most 52
         * significant bits, so a 64-bit integer or double-precision float type are safe for storing this identifier.
         *
         * @return int
         */
        public function getId(): int
        {
            return $this->id;
        }

        /**
         * True, if this user is a bot
         *
         * @return bool
         */
        public function isBot(): bool
        {
            return $this->is_bot;
        }

        /**
         * User's or bot's first name
         *
         * @return string
         */
        public function getFirstName(): string
        {
            return $this->first_name;
        }

        /**
         * Optional. User's or bot's last name
         *
         * @return string|null
         */
        public function getLastName(): ?string
        {
            return $this->last_name;
        }

        /**
         * Optional. User's or bot's username
         *
         * @return string|null
         */
        public function getUsername(): ?string
        {
            return $this->username;
        }

        /**
         * Optional. IETF language tag of the user's language
         *
         * @return string|null
         */
        public function getLanguageCode(): ?string
        {
            return $this->language_code;
        }

        /**
         * True, if this user is a Telegram Premium user
         * 
         * @return bool
         */
        public function isPremium(): bool
        {
            return $this->is_premium;
        }

        /**
         * Returns an array representation of the object.
         *
         * @param bool $raw
         * @return array
         */
        public function toArray(bool $raw = false): array
        {
            $return = [
                'id' => $this->id,
                'is_bot' => $this->is_bot,
                'first_name' => $this->first_name,
                'last_name' => $this->last_name,
                'username' => $this->username,
                'language_code' => $this->language_code,
                'is_premium' => $this->is_premium
            ];

            if(!$raw)
            {
                return $return;
            }

            return array_merge($return, [
                'updated_timestamp' => $this->updated_timestamp,
                'updated_client' => $this->updated_client
            ]);
        }

        /**
         * Constructs object from an array representation.
         *
         * @param array $array
         * @param bool $raw
         * @return TelegramUserMetadata
         */
        public static function fromArray(array $array, bool $raw = false): TelegramUserMetadata
        {
            $object = new self();
            
            $object->id = $array['id'];
            $object->is_bot = $array['is_bot'] ?? false;
            $object->first_name = $array['first_name'];
            $object->last_name = $array['last_name'] ?? null;
            $object->username = $array['username'] ?? null;
            $object->language_code = $array['language_code'] ?? null;
            $object->is_premium = $array['is_premium'] ?? false;

            if(!$raw)
            {
                return $object;
            }

            $object->updated_timestamp = $array['updated_timestamp'] ?? null;
            $object->updated_client = $array['updated_client'] ?? null;

            return $object;
        }
    }